import asyncio
import logging
import sys
from datetime import datetime

import httpx
from httpx import URL
from qasync import QEventLoop, asyncSlot
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio.session import async_sessionmaker as AsyncSessionMaker

import qcanvas.db as db
# noinspection PyUnresolvedReferences
import qcanvas.icons
from qcanvas.QtVersionHelper.QtCore import Signal
from qcanvas.QtVersionHelper.QtGui import QPixmap
from qcanvas.QtVersionHelper.QtWidgets import QApplication, QProgressDialog, QMainWindow
from qcanvas.net.canvas.canvas_client import CanvasClient
from qcanvas.ui.main_ui import AppMainWindow
from qcanvas.ui.setup_dialog import SetupDialog
from qcanvas.util import AppSettings
from qcanvas.util.constants import app_name
from qcanvas.util.course_indexer import DataManager
from qcanvas.util.linkscanner import CanvasFileScanner
from qcanvas.util.linkscanner.canvas_media_object_scanner import CanvasMediaObjectScanner
from qcanvas.util.linkscanner.dropbox_scanner import DropboxScanner

engine = create_async_engine("sqlite+aiosqlite:///canvas_db.😘", echo=False)

logging.basicConfig()
logging.getLogger("canvas_client").setLevel(logging.DEBUG)


async def begin():
    # Create meta stuff
    async with engine.begin() as conn:
        await conn.run_sync(db.Base.metadata.create_all)


class LoaderWindow(QMainWindow):
    init = Signal()
    setup = Signal()
    ready = Signal()

    def __init__(self):
        super().__init__()
        self.main_icon = QPixmap(":/main_icon.svg")
        self.init.connect(self.on_init)
        self.setup.connect(self.on_setup)
        self.ready.connect(self.on_ready)
        self.setWindowTitle(app_name)
        self.setWindowIcon(self.main_icon)
        self.setCentralWidget(QProgressDialog("Verifying config", None, 0, 0))

        self.init.emit()

    @asyncSlot()
    async def on_init(self):
        try:
            if not await CanvasClient.verify_config(AppSettings.canvas_url, AppSettings.canvas_api_key):
                self.setup.emit()
            else:
                self.ready.emit()
        except:
            self.setup.emit()

    @asyncSlot()
    async def on_setup(self):
        setup = SetupDialog(self)
        setup.setWindowIcon(self.main_icon)
        setup.rejected.connect(lambda: sys.exit(0))
        setup.accepted.connect(self.on_ready)
        setup.show()

    @asyncSlot()
    async def on_ready(self):
        client = CanvasClient(canvas_url=URL(AppSettings.canvas_url), api_key=AppSettings.canvas_api_key)
        data_manager = DataManager(
            client=client,
            link_scanners=[CanvasFileScanner(client), DropboxScanner(httpx.AsyncClient()),
                           CanvasMediaObjectScanner(client.client)],
            sessionmaker=AsyncSessionMaker(engine, expire_on_commit=False),
            last_update=datetime.min
        )

        await data_manager.init()
        self.close()
        self.open_main_app(data_manager)

    def open_main_app(self, data_manager: DataManager):
        self.main_window = AppMainWindow(data_manager)
        self.main_window.setWindowTitle(app_name)
        self.main_window.setWindowIcon(self.main_icon)
        self.main_window.show()
        # Set the main window as the parent of this window so this window is destroyed when the main window is closed
        self.setParent(self.main_window)


if __name__ == '__main__':
    asyncio.run(begin())

    app = QApplication(sys.argv)

    AppSettings.apply_selected_theme()

    event_loop = QEventLoop()
    asyncio.set_event_loop(event_loop)

    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set)

    loader_window = LoaderWindow()
    loader_window.show()

    with event_loop:
        event_loop.run_until_complete(app_close_event.wait())

    print("Exiting")



from qcanvas.QtVersionHelper.QtCore import QMimeDatabase
from qcanvas.QtVersionHelper.QtGui import QIcon
from qcanvas.QtVersionHelper.QtWidgets import QApplication, QStyle

_mime_database = QMimeDatabase()
_default_icon = None


def icon_for_filename(file_name: str) -> QIcon:
    global _default_icon

    for mime_type in _mime_database.mimeTypesForFileName(file_name):
        icon = QIcon.fromTheme(mime_type.iconName())

        if not icon.isNull():
            return icon

    if _default_icon is None:
        _default_icon = QApplication.style().standardIcon(QStyle.StandardPixmap.SP_FileIcon)

    return _default_icon



import logging
import sys
import traceback
from asyncio import Event
from typing import Sequence, Optional

from qasync import asyncSlot

import qcanvas.db.database as db
from qcanvas.QtVersionHelper.QtCore import Slot, Signal, Qt, QUrl
from qcanvas.QtVersionHelper.QtGui import QDesktopServices, QAction
from qcanvas.QtVersionHelper.QtWidgets import *
from qcanvas.ui.menu_bar.grouping_preferences_menu import GroupingPreferencesMenu
from qcanvas.ui.menu_bar.theme_selection_menu import ThemeSelectionMenu
from qcanvas.ui.viewer.course_list import CourseList
from qcanvas.ui.viewer.file_list import FileRow
from qcanvas.ui.viewer.file_view_tab import FileViewTab
from qcanvas.ui.viewer.page_list_viewer import AssignmentsViewer, PagesViewer, LinkTransformer
from qcanvas.util import AppSettings, self_updater
from qcanvas.util.constants import app_name
from qcanvas.util.course_indexer import DataManager

_aux_settings = AppSettings.auxiliary


class AppMainWindow(QMainWindow):
    logger = logging.getLogger()
    loaded = Signal()
    files_grouping_preference_changed = Signal(db.GroupByPreference)
    operation_lock = Event()

    def __init__(self, data_manager: DataManager, parent: QWidget | None = None):
        super().__init__(parent)

        self.group_by_pages_action: QAction | None = None
        self.group_by_modules_action: QAction | None = None
        self.file_grouping_menu: QMenu | None = None
        self.selected_course: db.Course | None = None
        self.courses: Sequence[db.Course] = []
        self.resources: dict[str, db.Resource] = {}
        self.data_manager = data_manager
        self.link_transformer = LinkTransformer(self.data_manager.link_scanners, self.resources)

        right_splitter = QSplitter()
        right_splitter.setOrientation(Qt.Orientation.Vertical)

        self.sync_button = QPushButton("Synchronize")
        self.sync_button.clicked.connect(self.sync_data)

        self.course_list = CourseList(self.data_manager)
        self.course_list.course_selected.connect(self.on_course_selected)

        self.assignment_viewer = AssignmentsViewer(self.link_transformer)
        self.assignment_viewer.viewer.anchorClicked.connect(self.viewer_link_clicked)

        self.pages_viewer = PagesViewer(self.link_transformer)
        self.pages_viewer.viewer.anchorClicked.connect(self.viewer_link_clicked)

        self.file_viewer = FileViewTab(data_manager.download_pool)
        self.file_viewer.files_column.tree.itemActivated.connect(self.download_file_from_file_pane)
        self.file_viewer.assignment_files_column.tree.itemActivated.connect(self.download_file_from_file_pane)

        self.tab_widget = QTabWidget()
        self.tab_widget.insertTab(0, self.file_viewer, "Files")
        self.tab_widget.insertTab(1, self.assignment_viewer, "Assignments")
        self.tab_widget.insertTab(2, self.pages_viewer, "Pages")

        h_layout = QHBoxLayout()
        h_layout.addWidget(self.course_list)
        h_layout.addWidget(self.tab_widget)
        h_layout.setStretch(1, 1)

        v_layout = QVBoxLayout()
        v_layout.addLayout(h_layout)
        v_layout.addWidget(self.sync_button)

        widget = QWidget()
        widget.setLayout(v_layout)
        self.setCentralWidget(widget)

        self.setup_menu_bar()

        self.files_grouping_preference_changed.connect(self.on_grouping_preference_changed)

        self.loaded.connect(self.load_course_list)
        self.loaded.connect(self.check_for_update)
        self.loaded.emit()

        self.restore_window_position()

    def setup_menu_bar(self):
        menu_bar = self.menuBar()

        menu_bar.addMenu(ThemeSelectionMenu())
        view_menu = menu_bar.addMenu("View")

        view_menu.addMenu(self.setup_group_by_menu())

    def setup_group_by_menu(self) -> QMenu:
        file_grouping_menu = GroupingPreferencesMenu(self.data_manager)
        self.course_list.course_selected.connect(file_grouping_menu.course_changed)
        file_grouping_menu.preference_changed.connect(self.on_grouping_preference_changed)

        return file_grouping_menu

    def closeEvent(self, event):
        _aux_settings.setValue("geometry", self.saveGeometry())
        _aux_settings.setValue("windowState", self.saveState())

    def restore_window_position(self):
        self.restoreGeometry(_aux_settings.value("geometry"))
        self.restoreState(_aux_settings.value("windowState"))

    @asyncSlot(QUrl)
    async def viewer_link_clicked(self, url: QUrl):
        # The url of a transformed link will start with an '@'
        if url.toString().startswith(LinkTransformer.transformed_url_prefix):
            # The rest of the 'url' is just the file id
            resource = self.resources[url.toString().removeprefix(LinkTransformer.transformed_url_prefix)]

            await self.data_manager.download_resource(resource)
            QDesktopServices.openUrl(QUrl.fromLocalFile(resource.download_location.absolute()))

    @asyncSlot(QTreeWidgetItem, int)
    async def download_file_from_file_pane(self, item: QTreeWidgetItem, _: int):
        if isinstance(item, FileRow):
            await self.data_manager.download_resource(item.resource)
            QDesktopServices.openUrl(QUrl.fromLocalFile(item.resource.download_location.absolute()))

    @asyncSlot()
    async def sync_data(self):
        # # self.operation_lock.
        self.sync_button.setEnabled(False)
        self.sync_button.setText("Synchronizing")
        try:
            await self.data_manager.synchronize_with_canvas()
            await self.load_course_list()

        finally:
            self.sync_button.setEnabled(True)
            self.sync_button.setText("Synchronize")

    @asyncSlot()
    async def load_course_list(self):
        self.courses = (await self.data_manager.get_data())
        self.selected_course = None
        self.resources.clear()

        for course in self.courses:
            self.resources.update({resource.id: resource for resource in course.resources})

        self.course_list.load_course_list(self.courses)

    @asyncSlot()
    async def check_for_update(self):
        try:
            newer_version, installed_version = await self_updater.get_newer_version()

            if newer_version is not None and newer_version != AppSettings.last_ignored_update:
                msg_box = QMessageBox(
                    QMessageBox.Icon.Question,
                    "Update available",
                    f"There is an update available ({installed_version} -> {newer_version})\nDo you want to update?\nThe program will close after the update is finished.",
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    self
                )

                def ignore_update():
                    AppSettings.last_ignored_update = newer_version

                msg_box.accepted.connect(self.do_self_update)
                msg_box.rejected.connect(ignore_update)
                msg_box.show()
            else:
                print("No update available (or skipping this update)")
        except BaseException as e:
            sys.stderr.write(f"Could not check for updates: {e}\n")
            sys.stderr.write("This can be ignored if in a dev environment\n")
            traceback.print_exc()

    @asyncSlot()
    async def do_self_update(self):
        try:
            progress_diag = QProgressDialog("Updating", None, 0, 0, self)
            progress_diag.setWindowTitle(app_name)
            progress_diag.show()
            await self_updater.do_update()
            self.close()
        except BaseException as e:
            traceback.print_exc()

            QMessageBox(
                QMessageBox.Icon.Critical,
                "Error",
                "An error occurred during the update",
                parent=self
            ).show()

    @Slot(db.Course)
    def on_course_selected(self, course: Optional[db.Course]):
        if course is not None:
            self.selected_course = course
            # todo these should really be slots connected to this signal...
            self.pages_viewer.fill_tree(course)
            self.assignment_viewer.fill_tree(course)
            self.file_viewer.load_course_files(course)
        else:
            self.selected_course = None
            self.file_viewer.clear()

    @asyncSlot(db.CoursePreferences)
    async def on_grouping_preference_changed(self):
        self.file_viewer.load_course_files(self.selected_course)



import asyncio
import json
import os
from importlib.metadata import version

import httpx
from packaging.version import Version

from qcanvas.util.constants import package_name


async def do_update():
    await asyncio.to_thread(os.system, f"pip install --upgrade {package_name}")


async def get_newer_version() -> tuple[Version | None, Version | None] | None:
    latest_version = await get_versions()
    installed_version = Version(version(package_name))

    print(f"latest = {latest_version}, installed = {installed_version}")

    if installed_version < latest_version:
        return latest_version, installed_version
    else:
        return None, None


async def get_versions() -> Version:
    async with httpx.AsyncClient() as client:
        data = json.loads((await client.get(f"https://pypi.org/pypi/{package_name}/json")).text)
        return Version(data["info"]["version"])



import qdarktheme

from qcanvas.QtVersionHelper.QtCore import QSettings, QUrl


def ensure_theme_is_valid(theme: str) -> str:
    if theme not in ["auto", "light", "dark", "native"]:
        return "light"
    else:
        return theme


class _AppSettings:
    def __init__(self):
        self.settings = QSettings("QCanvas", "client")
        self.auxiliary = QSettings("QCanvas", "ui")
        self._canvas_url = self.settings.value("canvas_url", None)
        self._api_key = self.settings.value("api_key", defaultValue=None)
        self._ignored_update = self.settings.value("ignored_update", defaultValue=None)
        self._theme = ensure_theme_is_valid(self.settings.value("theme", defaultValue="light"))

    @property
    def canvas_url(self) -> str | None:
        return self._canvas_url

    @canvas_url.setter
    def canvas_url(self, value: str):
        self._canvas_url = value
        self.settings.setValue("canvas_url", value)

    @property
    def canvas_api_key(self) -> str | None:
        return self._api_key

    @canvas_api_key.setter
    def canvas_api_key(self, value: str):
        self._api_key = value
        self.settings.setValue("api_key", value)

    @property
    def last_ignored_update(self) -> str | None:
        return self._ignored_update

    @last_ignored_update.setter
    def last_ignored_update(self, value: str):
        self._ignored_update = value
        self.settings.setValue("ignored_update", value)

    @property
    def theme(self) -> str | None:
        return self._theme

    @theme.setter
    def theme(self, value: str):
        value = ensure_theme_is_valid(value)
        self._theme = value
        self.settings.setValue("theme", value)

    # fixme should this really be here
    def apply_selected_theme(self):
        if self.theme != "native":
            qdarktheme.setup_theme(
                self.theme,
                custom_colors={"primary": "FF804F"}
            )

    @property
    def is_set(self):
        return self.canvas_url is not None and QUrl(self.canvas_url).isValid() and self.canvas_url is not None
